--
-- MulcherDensityUtil
--
--
-- @author John Deere 6930
-- @date 25/05/2021
-- @version 1.0
--

MulcherDensityUtil = {};

local DEBUG_MODE = false;

local PARAM_BETWEEN = "between";
local PARAM_GREATER = "greater";
local PARAM_EQUAL = "equal";
local MODIFIER_3POINTS = "ppp";
local MODIFIER_1POINT2VECTOR = "pvv";

local MulcherDensityUtil_mt = Class(MulcherDensityUtil);

MulcherDensityUtil.modDirectory = g_currentModDirectory;
MulcherDensityUtil.usePrecisionFarmingDLC = false;

MulcherDensityUtil.fruitDestructionModifiedNameToList = nil;

MulcherDensityUtil.PROCESSSTATE_NONE = 0;
MulcherDensityUtil.PROCESSSTATE_CHOPPEDSTRAW = 1;
MulcherDensityUtil.PROCESSSTATE_CHOPPEDSTRAWSPRAYLEVEL1 = 2;
MulcherDensityUtil.PROCESSSTATE_CHOPPEDSTRAWSPRAYLEVEL2 = 3;
MulcherDensityUtil.PROCESSSTATE_SPRAYLEVEL0 = 4;
MulcherDensityUtil.PROCESSSTATE_SPRAYLEVEL1 = 5;
MulcherDensityUtil.PROCESSSTATE_SPRAYLEVEL2 = 6;

function MulcherDensityUtil.initTerrain(terrainDetailId)
	local modifiers = {};
	modifiers.updateMulcherArea = {};
    modifiers.updateMulcherArea.modifier = DensityMapModifier:new(terrainDetailId, g_currentMission.sprayFirstChannel, g_currentMission.sprayNumChannels);
	modifiers.updateMulcherArea.fruitFilter = DensityMapFilter:new(modifiers.updateMulcherArea.modifier);
	modifiers.updateMulcherArea.filter1 = DensityMapFilter:new(modifiers.updateMulcherArea.modifier);
	modifiers.updateMulcherArea.filter2 = DensityMapFilter:new(modifiers.updateMulcherArea.modifier);
	modifiers.updateMulcherArea.maskFilter = DensityMapFilter:new(modifiers.updateMulcherArea.modifier);

    MulcherDensityUtil.densityMapModifiers = modifiers;
end;

-- Returns if fruit is modified by extern mod -> CropDestruction
-- @param string Name of fruit to proove
function MulcherDensityUtil.getIsFruitModified(name)
    if MulcherDensityUtil.fruitDestructionModifiedNameToList == nil then
        -- create simple table to look up for modified fruits
        -- needed to get correct cutState while using CropDestruction 2.0
        MulcherDensityUtil.fruitDestructionModifiedNameToList = {};
        for _, fruitType in pairs(g_fruitTypeManager.fruitTypes) do
            if MulcherDensityUtil.fruitDestructionModifiedNameToList[fruitType.name] == nil then
                if fruitType.destruction ~= nil and fruitType.destruction.state ~= nil and fruitType.destruction.state > (fruitType.cutState + 1) then
                    MulcherDensityUtil.fruitDestructionModifiedNameToList[fruitType.name] = true;
                else
                    MulcherDensityUtil.fruitDestructionModifiedNameToList[fruitType.name] = false;
                end;
            end;
        end;
    end;
    return MulcherDensityUtil.fruitDestructionModifiedNameToList[name] ~= nil and MulcherDensityUtil.fruitDestructionModifiedNameToList[name];
end;

-- UpdateMulcherArea
function MulcherDensityUtil.updateMulcherArea(fruitId, startWorldX, startWorldZ, widthWorldX, widthWorldZ, heightWorldX, heightWorldZ, setSprayLevel, setChoppedStraw, convertHeightDensityToSpray, worksInGround, prepareFruit, removeWeeds, eraseTireTracks, nitrogenAddFactor, limitToField)
    local currentProcessState = 0;
    local numChangedPixels = 0;    

	local fruit = g_currentMission.fruits[fruitId];
	local desc = g_fruitTypeManager:getFruitTypeByIndex(fruitId);
	if fruit == nil or fruit.id == 0 or desc.name == "WEED" then
		return 0;
	end;

    -- never prepare if tool works in ground or fruit is sugarCane
    if worksInGround or desc.name == "SUGARCANE" then
        prepareFruit = false;
    end;
    
	limitToField = Utils.getNoNil(limitToField, false);
    
	local modifiers = MulcherDensityUtil.densityMapModifiers.updateMulcherArea;
	local modifier = modifiers.modifier;
	local fruitFilter = modifiers.fruitFilter;
	local filter1 = modifiers.filter1;
	local filter2 = modifiers.filter2;
	local maskFilter = modifiers.maskFilter;
	modifier:setParallelogramWorldCoords(startWorldX, startWorldZ, widthWorldX, widthWorldZ, heightWorldX, heightWorldZ, MODIFIER_3POINTS);
	modifier:resetDensityMapAndChannels(fruit.id, desc.startStateChannel, desc.numStateChannels);

    local fruitFilterArea, prepareArea, lowerSprayArea, upperSprayArea = 0, 0, 0, 0;

	fruitFilter:resetDensityMapAndChannels(fruit.id, desc.startStateChannel, desc.numStateChannels);
	filter1:resetDensityMapAndChannels(fruit.id, desc.startStateChannel, desc.numStateChannels);
	filter2:resetDensityMapAndChannels(fruit.id, desc.startStateChannel, desc.numStateChannels);


    local cutStateAdder = 0;
    -- check for extern cutState modifiing
    if g_currentMission.fruitDestructionReloadedlist ~= nil then
        if MulcherDensityUtil.getIsFruitModified(desc.name) then
            cutStateAdder = 1;
        end;
    end;

    -- set fruitFilter for lower and upper
    local fruitFilterLower = 0;
    local fruitFilterUpper = desc.cutState + cutStateAdder + 1;
    if desc.preparedGrowthState > 0 then
        fruitFilterUpper = desc.preparedGrowthState + 1;
    end;
    if desc.regrows and not worksInGround then
        fruitFilterUpper = desc.cutState;
    end;
    local minFruitAvaiableState = desc.startStateChannel + 2;

    if worksInGround then
        -- inGround FruitFilter
        fruitFilterLower = desc.startStateChannel+1;
    else
        -- noGround FruitFilter
        fruitFilterLower = minFruitAvaiableState;
    end;
    -- minFruitAvaiableState === fruitFilterLower
    fruitFilter:setValueCompareParams(PARAM_BETWEEN, fruitFilterLower, fruitFilterUpper);

    fruitFilterArea, _ = modifier:executeGet(fruitFilter);
    if fruitFilterArea == 0 then
        return numChangedPixels;
    end;

    -- set filter for spray additions
    local sprayChangeState = desc.numGrowthStates/2;
    if desc.numGrowthStates%2 == 1 then
        sprayChangeState = (desc.numGrowthStates-1)/2;
    end;

    filter1:setValueCompareParams(PARAM_BETWEEN, minFruitAvaiableState, sprayChangeState);
    lowerSprayArea, _ = modifier:executeGet(filter1, fruitFilter);

    filter2:setValueCompareParams(PARAM_BETWEEN, sprayChangeState+1, desc.cutState);
    upperSprayArea, _ = modifier:executeGet(filter2, fruitFilter);

    local cutState = false;
    local prepareState = false;
    if lowerSprayArea == 0 and upperSprayArea == 0 then
        -- cut fruitFilter
        local cutArea = 0;
        if cutStateAdder > 0 then
            filter1:setValueCompareParams(PARAM_BETWEEN, desc.cutState, desc.cutState + cutStateAdder + 1);
            cutArea, _ = modifier:executeGet(filter1, fruitFilter);
        else
            filter1:setValueCompareParams(PARAM_EQUAL, desc.cutState+1);
            cutArea, _ = modifier:executeGet(filter1, fruitFilter);
        end;
        cutState = cutArea > 0;

        if desc.preparedGrowthState > 0 then
            -- prepare fruitFilter
            filter2:setValueCompareParams(PARAM_EQUAL, desc.preparedGrowthState+1);
            local preparedArea, _ = modifier:executeGet(filter2, fruitFilter);
            prepareState = preparedArea > 0;

            if prepareFruit and prepareState then
                return numChangedPixels;
            end;
        end;
    end;

    local executeSet = false;
    local executeVal = 0;
    
    if worksInGround then
        -- tool works in ground
        -- reset fruit (ground is able to get fruit haulm)
        executeSet = true;
        executeVal = desc.cutState + cutStateAdder + 3;
    else
        -- tool doesnt work in ground, is able to set cutState
        if prepareFruit then
            -- tool is able to prepare fruit
            if desc.preparedGrowthState > 0 then
                -- check for preparing
                if desc.minPreparingGrowthState >= 0 and desc.maxPreparingGrowthState >= 0 then
                    filter2:setValueCompareParams(PARAM_BETWEEN, desc.minPreparingGrowthState + 1, desc.maxPreparingGrowthState + 1)
                    prepareArea, _ = modifier:executeGet(filter2);
                end;
                
                if prepareArea > 0 then
                    -- prepareFruit
                    FSDensityMapUtil.updateFruitPreparerArea(fruitId, startWorldX, startWorldZ, widthWorldX, widthWorldZ, heightWorldX, heightWorldZ, startWorldX, startWorldZ, widthWorldX, widthWorldZ, heightWorldX, heightWorldZ);
                else
                    -- reset fruit, not prepare ready
                    executeSet = true;
                    executeVal = desc.cutState + 1;
                end;
            else
                -- no preparing possible
                executeSet = true;
                if desc.regrows then
                    executeVal = desc.cutState + 1;
                else
                    executeVal = desc.cutState + cutStateAdder + 2;
                end;
            end;
        else
            -- tool is not able to prepare fruit
            executeSet = true;
            if desc.regrows then
                executeVal = desc.cutState + 1;
            else
                executeVal = desc.cutState + cutStateAdder + 3;
            end;
        end;
    end;

    if executeSet then
        if limitToField then
            maskFilter:resetDensityMapAndChannels(g_currentMission.terrainDetailId, g_currentMission.terrainDetailTypeFirstChannel, g_currentMission.terrainDetailTypeNumChannels);
            maskFilter:setValueCompareParams(PARAM_GREATER, 0);
            
            _, numChangedPixels = modifier:executeSet(executeVal, fruitFilter, maskFilter);
        else
            _, numChangedPixels = modifier:executeSet(executeVal, fruitFilter);
        end;
    end;

    -- calc choppedStraw and SprayLevel
    if prepareArea <= 0 or not prepareFruit and prepareArea > 0 then
        if cutState or prepareState then
            if not setChoppedStraw then
                currentProcessState = MulcherDensityUtil.PROCESSSTATE_SPRAYLEVEL0;
            else
                currentProcessState = MulcherDensityUtil.PROCESSSTATE_CHOPPEDSTRAW;
            end;
        end;
        if lowerSprayArea > 0 then
            if not setChoppedStraw then
                currentProcessState = MulcherDensityUtil.PROCESSSTATE_SPRAYLEVEL1;
            else
                currentProcessState = MulcherDensityUtil.PROCESSSTATE_CHOPPEDSTRAWSPRAYLEVEL1;
            end;
        end;
        if upperSprayArea > 0 then
            if not setChoppedStraw then
                currentProcessState = MulcherDensityUtil.PROCESSSTATE_SPRAYLEVEL2;
            else
                currentProcessState = MulcherDensityUtil.PROCESSSTATE_CHOPPEDSTRAWSPRAYLEVEL2;
            end;
        end;
    end;

    -- clearHeightMap
    if convertHeightDensityToSpray then
        local fillTypeIndex = DensityMapHeightUtil.getFillTypeAtArea(startWorldX, startWorldZ, widthWorldX, widthWorldZ, heightWorldX, heightWorldZ);
        if fillTypeIndex ~= FillType.UNKNOWN then
            if currentProcessState ~= MulcherDensityUtil.PROCESSSTATE_NONE then
                if currentProcessState >= MulcherDensityUtil.PROCESSSTATE_SPRAYLEVEL0 then
                    if not setChoppedStraw then
                        currentProcessState = math.min(currentProcessState+1, MulcherDensityUtil.PROCESSSTATE_SPRAYLEVEL2);
                    else
                        currentProcessState = math.min(currentProcessState+1, MulcherDensityUtil.PROCESSSTATE_CHOPPEDSTRAWSPRAYLEVEL2);
                    end;
                end;
            end;
        end;
    end;
    DensityMapHeightUtil.clearArea(startWorldX, startWorldZ, widthWorldX, widthWorldZ, heightWorldX, heightWorldZ);

    if currentProcessState ~= MulcherDensityUtil.PROCESSSTATE_NONE then
        if MulcherDensityUtil.usePrecisionFarmingDLC then
            -- precisionFarming DLC is used!
            nitrogenAddFactor = Utils.getNoNil(nitrogenAddFactor, 0.0);
            local nitrogenMap = FS19_precisionFarming.g_precisionFarming.nitrogenMap;
            if nitrogenMap ~= nil then
                local sprayAmountManual = 0;
                local forcedGroundType = SprayType.FERTILIZER;

                if currentProcessState == MulcherDensityUtil.PROCESSSTATE_CHOPPEDSTRAW then
                    if fruit.preparingOutputId ~= 0 then
                        modifier:resetDensityMapAndChannels(fruit.preparingOutputId, 0, 1);
                        filter2:setValueCompareParams(PARAM_GREATER, 0);
                        modifier:executeSet(1, filter2);
                    else
                        forcedGroundType = g_currentMission.chopperGroundLayerType;
                    end;
                    sprayAmountManual = 1;
                elseif currentProcessState == MulcherDensityUtil.PROCESSSTATE_CHOPPEDSTRAWSPRAYLEVEL1 then
                    if fruit.preparingOutputId ~= 0 then
                        modifier:resetDensityMapAndChannels(fruit.preparingOutputId, 0, 1);
                        filter2:setValueCompareParams(PARAM_GREATER, 0);
                        modifier:executeSet(1, filter2);
                    else
                        forcedGroundType = g_currentMission.chopperGroundLayerType;
                    end;
                    sprayAmountManual = 2;
                elseif currentProcessState == MulcherDensityUtil.PROCESSSTATE_CHOPPEDSTRAWSPRAYLEVEL2 then
                    if fruit.preparingOutputId ~= 0 then
                        modifier:resetDensityMapAndChannels(fruit.preparingOutputId, 0, 1);
                        filter2:setValueCompareParams(PARAM_GREATER, 0);
                        modifier:executeSet(1, filter2);
                    else
                        forcedGroundType = g_currentMission.chopperGroundLayerType;
                    end;
                    sprayAmountManual = 3;
                elseif currentProcessState == MulcherDensityUtil.PROCESSSTATE_SPRAYLEVEL0 then
                    -- do nothing
                elseif currentProcessState == MulcherDensityUtil.PROCESSSTATE_SPRAYLEVEL1 then
                    sprayAmountManual = 2;
                elseif currentProcessState == MulcherDensityUtil.PROCESSSTATE_SPRAYLEVEL2 then
                    sprayAmountManual = 3;
                else
                    print("MulcherDensityUtil: Something wrong happened :( ")
                end;

                if sprayAmountManual > 0 then
                    local modifiedSprayType = g_sprayTypeManager:getSprayTypeByIndex(SprayType.FERTILIZER);
                    if modifiedSprayType ~= nil then
                        -- Set GroundLockType of FERTILIZER
                        local oldGroundType = modifiedSprayType.groundType;
                        if modifiedSprayType.groundType ~= forcedGroundType then
                            modifiedSprayType.groundType = forcedGroundType;
                        end;
                        sprayAmountManual = sprayAmountManual + nitrogenAddFactor;
                        
                        nitrogenMap:updateSprayArea(startWorldX, startWorldZ, widthWorldX, widthWorldZ, heightWorldX, heightWorldZ, SprayType.FERTILIZER, SprayType.FERTILIZER, false, sprayAmountManual, nil);
                        FSDensityMapUtil.setGroundTypeLayerArea(startWorldX, startWorldZ, widthWorldX, widthWorldZ, heightWorldX, heightWorldZ, forcedGroundType);
                        
                        -- Reset GroundLockType of FERTILIZER
                        modifiedSprayType.groundType = oldGroundType;
                    end;
                end;
            end;
        else
            -- precisionFarming DLC is not used!
            if currentProcessState == MulcherDensityUtil.PROCESSSTATE_CHOPPEDSTRAW then
                if fruit.preparingOutputId ~= 0 then
                    modifier:resetDensityMapAndChannels(fruit.preparingOutputId, 0, 1);
                    filter2:setValueCompareParams(PARAM_GREATER, 0);
                    modifier:executeSet(1, filter2);
                else
                    FSDensityMapUtil.setGroundTypeLayerArea(startWorldX, startWorldZ, widthWorldX, widthWorldZ, heightWorldX, heightWorldZ, g_currentMission.chopperGroundLayerType);
                end;
            elseif currentProcessState == MulcherDensityUtil.PROCESSSTATE_CHOPPEDSTRAWSPRAYLEVEL1 then
                if fruit.preparingOutputId ~= 0 then
                    FSDensityMapUtil.updateFertilizerArea(startWorldX, startWorldZ, widthWorldX, widthWorldZ, heightWorldX, heightWorldZ, SprayType.FERTILIZER);
                    modifier:resetDensityMapAndChannels(fruit.preparingOutputId, 0, 1);
                    filter2:setValueCompareParams(PARAM_GREATER, 0);
                    modifier:executeSet(1, filter2);
                else
                    local setsStraw = FSDensityMapUtil.updateFertilizerArea(startWorldX, startWorldZ, widthWorldX, widthWorldZ, heightWorldX, heightWorldZ, g_currentMission.chopperGroundLayerType);
                    if setsStraw == 0 then
                        FSDensityMapUtil.setGroundTypeLayerArea(startWorldX, startWorldZ, widthWorldX, widthWorldZ, heightWorldX, heightWorldZ, g_currentMission.chopperGroundLayerType);
                    end;
                end;
            elseif currentProcessState == MulcherDensityUtil.PROCESSSTATE_CHOPPEDSTRAWSPRAYLEVEL2 then
                FSDensityMapUtil.updateSprayArea(startWorldX, startWorldZ, widthWorldX, widthWorldZ, heightWorldX, heightWorldZ, SprayType.FERTILIZER);
                if fruit.preparingOutputId ~= 0 then
                    FSDensityMapUtil.updateFertilizerArea(startWorldX, startWorldZ, widthWorldX, widthWorldZ, heightWorldX, heightWorldZ, SprayType.HERBICIDE);
                    modifier:resetDensityMapAndChannels(fruit.preparingOutputId, 0, 1);
                    filter2:setValueCompareParams(PARAM_GREATER, 0);
                    modifier:executeSet(1, filter2);
                else
                    local setsStraw = FSDensityMapUtil.updateFertilizerArea(startWorldX, startWorldZ, widthWorldX, widthWorldZ, heightWorldX, heightWorldZ, g_currentMission.chopperGroundLayerType);
                    if setsStraw == 0 then
                        FSDensityMapUtil.setGroundTypeLayerArea(startWorldX, startWorldZ, widthWorldX, widthWorldZ, heightWorldX, heightWorldZ, g_currentMission.chopperGroundLayerType);
                    end;
                end;
            elseif currentProcessState == MulcherDensityUtil.PROCESSSTATE_SPRAYLEVEL0 then
                -- do nothing
            elseif currentProcessState == MulcherDensityUtil.PROCESSSTATE_SPRAYLEVEL1 then
                FSDensityMapUtil.updateFertilizerArea(startWorldX, startWorldZ, widthWorldX, widthWorldZ, heightWorldX, heightWorldZ, SprayType.FERTILIZER);
            elseif currentProcessState == MulcherDensityUtil.PROCESSSTATE_SPRAYLEVEL2 then
                FSDensityMapUtil.updateSprayArea(startWorldX, startWorldZ, widthWorldX, widthWorldZ, heightWorldX, heightWorldZ, SprayType.FERTILIZER);
                FSDensityMapUtil.updateFertilizerArea(startWorldX, startWorldZ, widthWorldX, widthWorldZ, heightWorldX, heightWorldZ, 0);
            else
                print("MulcherDensityUtil: Something wrong happened :( ")
            end;
        end;
    end;

    -- DEBUG_MODE
    if DEBUG_MODE then
        local dPrint = "";
        if fruitFilterArea ~= nil and fruitFilterArea ~= 0 then
            dPrint = string.format("%s%s", dPrint, desc.name);
        end;

         if fruitFilterArea ~= nil and fruitFilterArea ~= 0 then dPrint = dPrint.." [fruitFilterArea changed]"; end;
         if lowerSprayArea ~= nil and lowerSprayArea ~= 0 then dPrint = dPrint.." [lowerSprayArea changed]"; end;
         if upperSprayArea ~= nil and upperSprayArea ~= 0 then dPrint = dPrint.." [upperSprayArea changed]"; end;
         if prepareArea ~= nil and prepareArea ~= 0 then dPrint = dPrint.." [prepareArea changed]"; end;
         if cutState then dPrint = dPrint.." [cutState]"; end;
        -- if prepareState then dPrint = dPrint.." [prepareState]"; end;
        if currentProcessState ~= nil then dPrint = dPrint.." [currentProcessState_"..currentProcessState.."]"; end;

        if dPrint ~= "" then print(dPrint); end;
    end;

    -- remove weeds
    if removeWeeds and g_currentMission.missionInfo.weedsEnabled then
        FSDensityMapUtil.removeWeedArea(startWorldX, startWorldZ, widthWorldX, widthWorldZ, heightWorldX, heightWorldZ);
    end;

    -- erase TireTracks
    if eraseTireTracks then
        FSDensityMapUtil.eraseTireTrack(startWorldX, startWorldZ, widthWorldX, widthWorldZ, heightWorldX, heightWorldZ);
    end;

	return numChangedPixels;
end;

-------------------------------
--------Initialization---------
-------------------------------
function init()
	FSDensityMapUtil.initTerrain = Utils.appendedFunction(FSDensityMapUtil.initTerrain, initTerrain);
    FSBaseMission.loadMapFinished = Utils.appendedFunction(FSBaseMission.loadMapFinished, onLoadMapFinished)
end;

function onLoadMapFinished(mission, node)
    if g_modIsLoaded["FS19_precisionFarming"] then
        -- is precisionFarming loaded?
        MulcherDensityUtil.usePrecisionFarmingDLC = _G["FS19_precisionFarming"] ~= nil and _G["FS19_precisionFarming"].g_precisionFarming ~= nil;
    end;
end;

function initTerrain(terrainDetailId)
    MulcherDensityUtil.initTerrain(terrainDetailId);
end;

init();